#!/usr/bin/env python

#===============================================================================
#    ProVerif Editor
#    Copyright (C) 2009 Joeri de Ruiter
# 
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
# 
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
# 
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#===============================================================================

#TODO: Backups when saving

from parser import Parser, ParseException, ParseDeclarationException
from tokens import *
from xml.dom import minidom
import gtk
import gtksourceview2
import pygtk
import re
import subprocess
import tempfile
pygtk.require('2.0')


class Editor:
    def __init__(self):
        # Load GUI from file
        builder = gtk.Builder()
        self.builder = builder
        builder.add_from_file('editor.ui')
        
        # Get the window widget
        self.window = builder.get_object('window')

        # Connect events of window widget
        self.window.connect("delete_event", self.delete_event)
        self.window.connect("destroy", self.destroy)        
        
        self.window.set_title('ProVerif Editor - Untitled')
        
        # Add parser
        self.parser = Parser()
        
        # Add accelerators
        accel_group = gtk.AccelGroup()
        self.window.add_accel_group(accel_group)
        builder.get_object('new').add_accelerator("activate", accel_group, ord("N"), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        builder.get_object('save').add_accelerator("activate", accel_group, ord("S"), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        builder.get_object('undo').add_accelerator("activate", accel_group, ord("Z"), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        builder.get_object('redo').add_accelerator("activate", accel_group, ord("Y"), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        builder.get_object('cut').add_accelerator("activate", accel_group, ord("X"), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        builder.get_object('copy').add_accelerator("activate", accel_group, ord("C"), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        builder.get_object('paste').add_accelerator("activate", accel_group, ord("V"), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        builder.get_object('comment').add_accelerator("activate", accel_group, ord("D"), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
            
        # Initialize templates menu
        templates_menu = gtk.Menu()
        builder.get_object('menuitem_templates').set_submenu(templates_menu)
        
        # Load and parse xml file
        xmldoc = minidom.parse('templates.xml')
        xml_templates = xmldoc.getElementsByTagName('template')
        for template in xml_templates:
            name = template.getElementsByTagName('name')[0].firstChild.data
            value = template.getElementsByTagName('value')[0].firstChild.data.replace('\\n', '\n')
            item = gtk.MenuItem(name)
            item.connect('activate', self.do_insert, value)
            templates_menu.append(item)
        
        templates_menu.show_all()

        
        # Initialize index treeview
        self.treeview_index = builder.get_object('treeview_index')
        
        column = gtk.TreeViewColumn()
        self.treeview_index.append_column(column)
        cell = gtk.CellRendererText()
        column.pack_start(cell, True)
        column.add_attribute(cell, 'text', 0)
        column.set_resizable(True)
        
        self.treestore_index = builder.get_object('treestore_index')
        self.treestore_index_functions = self.treestore_index.append(None, ['Functions'])
        self.treestore_index_data = self.treestore_index.append(None, ['Data'])
        self.treestore_index_reducs = self.treestore_index.append(None, ['Reductions'])
        self.treestore_index_processes = self.treestore_index.append(None, ['Processes'])        

        self.treeview_index.get_selection().connect("changed", self.select_index)

        # Initialize results treeview
        self.treeview_results = builder.get_object('treeview_results')
        
        column = gtk.TreeViewColumn()
        self.treeview_results.append_column(column)
        cell = gtk.CellRendererText()
        column.pack_start(cell, True)
        column.add_attribute(cell, 'text', 0)        

        self.liststore_results = builder.get_object('liststore_results')

        # Initialize warnings treeview
        self.treeview_warnings = builder.get_object('treeview_warnings')
        
        column = gtk.TreeViewColumn()
        self.treeview_warnings.append_column(column)
        cell = gtk.CellRendererText()
        column.pack_start(cell, True)
        column.add_attribute(cell, 'text', 0)        
        
        self.liststore_warnings = builder.get_object('liststore_warnings')
        
        self.treeview_warnings.get_selection().connect("changed", self.update_message)

        # Initialize errors treeview
        self.treeview_errors = builder.get_object('treeview_errors')
        
        column = gtk.TreeViewColumn()
        self.treeview_errors.append_column(column)
        cell = gtk.CellRendererText()
        column.pack_start(cell, True)
        column.add_attribute(cell, 'text', 0)        
        
        self.liststore_errors = builder.get_object('liststore_errors')

        self.treeview_errors.get_selection().connect("changed", self.update_message)
       
        # Initialize full results treeview
        self.treeview_results_full = builder.get_object('treeview_results_full')
        
        column = gtk.TreeViewColumn()
        self.treeview_results_full.append_column(column)
        cell = gtk.CellRendererText()
        column.pack_start(cell, True)
        column.add_attribute(cell, 'text', 0)        

        self.treestore_results = builder.get_object('treestore_results')
       
       
        # Add references to the following UI elements
        self.eventbox_parse_error = builder.get_object('eventbox_parse_error')
        self.label_parse_error = builder.get_object('label_parse_error')
        
        self.notebook_messages = builder.get_object('notebook_messages')
        self.scrolledwindow_results = builder.get_object('scrolledwindow_results')
        self.scrolledwindow_warnings = builder.get_object('scrolledwindow_warnings')
        self.scrolledwindow_errors = builder.get_object('scrolledwindow_errors')
        
        self.textview_output = builder.get_object('textview_output')
        self.statusbar = builder.get_object('statusbar')
        self.aboutdialog = builder.get_object('aboutdialog')
        
        # Connect signals to methods
        signals = {
                  'on_new': self.do_new,
                  'on_open': self.do_open,
                  'on_save': self.do_save,
                  'on_save_as': self.do_save_as,
                  'on_quit': self.delete_event,
                  'on_cut': self.do_cut,
                  'on_copy': self.do_copy,
                  'on_paste': self.do_paste,
                  'on_undo': self.do_undo,
                  'on_redo': self.do_redo,
                  'on_comment': self.do_comment,
                  'on_execute': self.do_execute,
                  'on_about': self.show_about}
        builder.connect_signals(signals)

        #TODO Move Buffer and sourceview widget to GUI file
        # Create buffer        
        self.buffer = Buffer()
        
        # Add tags to buffer
        self.texttag_error = builder.get_object('texttag_error')
        self.buffer.get_tag_table().add(self.texttag_error)
        
        self.texttag_warning = builder.get_object('texttag_warning')
        self.buffer.get_tag_table().add(self.texttag_warning)
        
        self.texttag_parse_error = builder.get_object('texttag_parse_error')
        self.buffer.get_tag_table().add(self.texttag_parse_error)
        
        self.texttag_declaration = builder.get_object('texttag_declaration')
        self.buffer.get_tag_table().add(self.texttag_declaration)
        
        self.texttag_reference = builder.get_object('texttag_reference')
        self.buffer.get_tag_table().add(self.texttag_reference)
        
        #Create sourceview widget
        self.view = gtksourceview2.View(self.buffer)
        self.view.set_show_line_numbers(True)
        self.view.get_buffer().set_highlight_syntax(True)
        
        # Connect buffer events
        self.buffer.connect("changed", self.update_undo)
        #TODO Move into Buffer class
        self.buffer.connect('changed', self.highlight_code)
        self.buffer.connect("mark_set", self.update_context_highlighting)
        
        # Load spi language file for highlighting
        manager = gtksourceview2.LanguageManager();
        manager.set_search_path(manager.get_search_path() + ['./'])
        self.view.get_buffer().set_language(manager.get_language('spi'))
        
        # Add sourceview widget to window
        sw_edit = builder.get_object('scrolledwindow_edit')
        sw_edit.add(self.view)        
        
        # Show all widgets in the window
        self.window.show_all()     
        self.update_undo()
    
    def main(self):
        gtk.main()
        
    def check_modified(self):
        # Check whether the fileis modified
        if self.view.get_buffer().get_modified():
            # If the file is modified, ask whether changes should be lost
            err = 'File is modified. Close file without saving?'
            dialog = gtk.MessageDialog(self.window, gtk.DIALOG_MODAL,
						gtk.MESSAGE_QUESTION,
						gtk.BUTTONS_YES_NO, err);
            result = dialog.run()
            dialog.destroy()

            # Return False if the user does not want to close the file
            if result != gtk.RESPONSE_YES:
                return False

        # Return True indicating the file can be closed
        return True
    

    def delete_event(self, widget, event=None, data=None):
        # Check whether the text is modified
        if self.check_modified():
            gtk.main_quit()        
        # Return True if the window cannot be closed
        return True

    def destroy(self, widget, data=None):
        # Destroy the window
        gtk.main_quit()

    def do_comment(self, widget):
        # Some text has to be selected to be commented
        if not self.buffer.get_has_selection():
            return
        
        # Begin user action to allow undo/redo
        self.buffer.begin_user_action()
        
        # Add start of comment
        (start, end) = self.buffer.get_selection_bounds()
        self.buffer.insert_interactive(start, '(*', self.view.get_editable())
        
        # Add end of comment
        (start, end) =  self.buffer.get_selection_bounds()
        self.buffer.insert_interactive(end, '*)', self.view.get_editable())
        
        # End user action so the complete action of commenting is seen as one action
        self.buffer.end_user_action()
        
        
        self.buffer.place_cursor(end)
        
        
    def do_new(self, widget):
        # Check whether the text is modified and can be closed
        if self.check_modified():
            # Reset buffer
            self.buffer.set_filename(None)
            self.buffer.clear()
            
            self.update_undo()
            
            # Clear output view
            self.textview_output.get_buffer().set_text('')
            
            # Clear all messages
            self.reset_messages()
            
            # Update window title and status information
            self.window.set_title('ProVerif Editor - Untitled')
            context = self.statusbar.get_context_id("Status")
            self.statusbar.push(context, "New file")

    def do_open(self, widget):
        # Check whether the text is modified and can be closed
        if self.check_modified():
            # Get a filename from the user using the file chooser dialog 
            dialog = gtk.FileChooserDialog("Open", self.window, gtk.FILE_CHOOSER_ACTION_OPEN, (gtk.STOCK_OPEN, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT))
            result = dialog.run()
            filename = dialog.get_filename()
            dialog.destroy()

            if result == gtk.RESPONSE_ACCEPT:
                # Try to open the given file
                self.buffer.set_filename(filename)
                result = self.buffer.load()
                
                self.update_undo()
                
                if result:
                    # Clear output view
                    self.textview_output.get_buffer().set_text('')
                    
                    # Clear all messages
                    self.reset_messages()
                    
                    # Update window title and status information
                    self.window.set_title('ProVerif Editor - ' + filename)
                    context = self.statusbar.get_context_id("Status")
                    self.statusbar.push(context, "File opened")
                    
    def do_save(self, widget):
        # If no filename is set yet, open the Save as.. dialog
        if not self.buffer.filename:
            return self.do_save_as(widget)
        else:
            # Try to save to the file
            result = self.buffer.save()
            if result:
                # Update status information
                context = self.statusbar.get_context_id("Status")
                self.statusbar.push(context, "File saved")
                return True
        return False

    def do_save_as(self, widget):
        # Get a filename from the user using the file chooser dialog 
        dialog = gtk.FileChooserDialog("Save as...", self.window, gtk.FILE_CHOOSER_ACTION_SAVE, (gtk.STOCK_SAVE, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT))
        result = dialog.run()
        filename = dialog.get_filename()
        dialog.destroy()

        if result == gtk.RESPONSE_ACCEPT:
            # Try to save to the given filename
            self.buffer.set_filename(filename)
            result = self.buffer.save()
            
            if result:
                # Update status information and window title
                context = self.statusbar.get_context_id("Status")
                self.statusbar.push(context, "File saved")
                self.window.set_title('ProVerif Editor - ' + filename)
                return True
        return False

    def do_execute(self, widget):
        if self.buffer.get_modified():
            # If the file is modified, save to a temporary file so we can still check it using ProVerif
            #TODO This does not work on Windows
            file = tempfile.NamedTemporaryFile()
            filename = file.name
            self.buffer.save(filename)
        else:
            filename = self.view.get_buffer().filename
        
        # Update status information
        context = self.statusbar.get_context_id("Status")
        self.statusbar.push(context, "ProVerif executing...")

        # Execute ProVerif adn retrieve output
        output = subprocess.Popen(["proverif", "-in", "pi", filename], stdout=subprocess.PIPE).communicate()[0]
        
        # Display complete output in output tab
        self.textview_output.get_buffer().set_text(output)

        # Update status information
        context = self.statusbar.get_context_id("Status")
        self.statusbar.push(context, "ProVerif executed")

        # Initialize parsing
        last_pos = None
        self.reset_messages()
        self.treestore_results.clear()
        process = False
        process_list = []
        trace = False
        trace_list = []
        results = 0
        warnings = 0
        errors = 0

        # Parse output line by line
        for line in output.splitlines():
            if errors > 0:
                self.liststore_errors.append([line, last_pos['start_line'], last_pos['start_char'], last_pos['end_line'], last_pos['end_char']])
            elif line.startswith('-- '):
                process = False
                trace = True
                trace_list.append([])
                trace_list[len(trace_list) - 1].append(line)
            elif line.startswith('File'):
                m = re.match(r"File \"([\S\ ]+)\", line (\d+), character (\d+) - line (\d+), character (\d+)", line)
                if not m == None:
                    last_pos = {'filename': m.group(1), 'start_line': int(m.group(2)) - 1, 'start_char': int(m.group(3)) - 1, 'end_line': int(m.group(4)) - 1, 'end_char': int(m.group(5)) - 1}
            elif line.startswith('Warning:'):
                warnings += 1
                self.liststore_warnings.append([line[9:len(line)], last_pos['start_line'], last_pos['start_char'], last_pos['end_line'], last_pos['end_char']])
            elif line.startswith('Error:'):
                errors += 1
                self.liststore_errors.append([line[7:len(line)], last_pos['start_line'], last_pos['start_char'], last_pos['end_line'], last_pos['end_char']])
            elif trace:
                trace_list[len(trace_list) - 1].append(line)
                
                if line.startswith('RESULT '):
                    results += 1
                    
                    self.liststore_results.append([line[7:len(line)]])         
                    parent = self.treestore_results.append(None, [line])
                    rules = self.treestore_results.append(parent, ['Rules'])
                    
                    trace_lines = trace_list[len(trace_list)-1]
                    
                    if trace_lines[0] == '-- Secrecy & events.':
                        rule = []
                        cnt = 2;
                        for trace_line in trace_lines[2:]:
                            if trace_line.startswith('Rule '):
                                if len(rule) > 0:
                                    rule_row = self.treestore_results.append(rules, [rule[0]])
                                    self.treestore_results.append(rule_row, ['\n'.join(rule[1:])])
                                rule = []
                                
                            if not trace_line == '':
                                rule.append(trace_line)
                            
                            cnt += 1
                            if trace_line == 'Completing...':
                                break
                            
                        self.treestore_results.append(parent, ['\n'.join(trace_list[len(trace_list)-1][cnt:])])
                    else:
                        self.treestore_results.append(parent, ['\n'.join(trace_list[len(trace_list)-1])])
            elif process:
                process_list.append(line)
            elif line == 'Process:':
                process = True
                
        if len(process_list) > 0:
            parent = self.treestore_results.prepend(None, ['Process'])
            self.treestore_results.append(parent, ['\n'.join(process_list)])
                    
        # Update the title of the message tabs so they contain the number of messages
        self.notebook_messages.set_tab_label(self.scrolledwindow_results, gtk.Label('Results (' + str(results) + ')'))
        self.notebook_messages.set_tab_label(self.scrolledwindow_warnings, gtk.Label('Warnings (' + str(warnings) + ')'))
        self.notebook_messages.set_tab_label(self.scrolledwindow_errors, gtk.Label('Errors (' + str(errors) + ')'))

    def do_insert(self, widget, str = None):
        # Dirty hack to be able to use this method both when using events and when calling it directly
        if str == None:
            str = widget

        # Insert text interactively so users can use undo to remove it again
        self.buffer.insert_interactive_at_cursor(str, self.view.get_editable())

    def do_undo(self, widget):
        # Check whether it is possible to undo
        if self.buffer.can_undo():
            self.buffer.undo()
            
        # Update undo status
        self.update_undo()

    def do_redo(self, widget):
        # Check whether it is possible to redo
        if self.buffer.can_redo():
            self.buffer.redo()
            
        # Update undo status
        self.update_undo()

    def update_undo(self, widget = None):
        # Enable/disable menu items for undo/redo depending on whether they are possible
        self.builder.get_object('undo').set_sensitive(self.buffer.can_undo())
        self.builder.get_object('redo').set_sensitive(self.buffer.can_redo())

    def do_cut(self, widget):
        self.buffer.cut_clipboard(gtk.Clipboard(), self.view.get_editable())

    def do_copy(self, widget):
        self.buffer.copy_clipboard(gtk.Clipboard())

    def do_paste(self, widget):
        self.buffer.paste_clipboard(gtk.Clipboard(), None, self.view.get_editable())

    def show_about(self, widget):
        # Show about dialog
        self.aboutdialog.run()
        self.aboutdialog.hide()

    def update_message(self, selection):
        buffer = self.view.get_buffer()

        if not selection.get_selected()[1] == None:
            self.reset_tags()
            self.treeview_index.get_selection().unselect_all()
            start_line = int(selection.get_selected()[0].get_value(selection.get_selected()[1], 1))
            start_char = int(selection.get_selected()[0].get_value(selection.get_selected()[1], 2))
            end_line = int(selection.get_selected()[0].get_value(selection.get_selected()[1], 3))
            end_char = int(selection.get_selected()[0].get_value(selection.get_selected()[1], 4))
            buffer.apply_tag(self.texttag_error, buffer.get_iter_at_line_offset(start_line, start_char), buffer.get_iter_at_line_offset(end_line, end_char))


    def reset_tags(self, widget = None):
        # Get start and end position of the buffer
        start = self.buffer.get_start_iter()
        end = self.buffer.get_end_iter()
        
        # Remove all tags from the buffer
        self.buffer.remove_tag(self.texttag_error, start, end)
        self.buffer.remove_tag(self.texttag_warning, start, end)
        self.buffer.remove_tag(self.texttag_declaration, start, end)
        self.buffer.remove_tag(self.texttag_reference, start, end)

    def reset_selections(self):
        # Unselect all messages
        self.treeview_warnings.get_selection().unselect_all()
        self.treeview_errors.get_selection().unselect_all()
        self.treeview_index.get_selection().unselect_all()
    
    def reset_index(self):
        # Clear the index
        self.remove_children(self.treestore_index_functions)
        self.remove_children(self.treestore_index_data)
        self.remove_children(self.treestore_index_reducs)
        self.remove_children(self.treestore_index_processes)

    def reset_messages(self):
        # Clear all messages
        self.liststore_results.clear()
        self.liststore_warnings.clear()
        self.liststore_errors.clear()
        
        # Reset the title of the tabs
        self.notebook_messages.set_tab_label(self.scrolledwindow_results, gtk.Label('Results'))
        self.notebook_messages.set_tab_label(self.scrolledwindow_warnings, gtk.Label('Warnings'))
        self.notebook_messages.set_tab_label(self.scrolledwindow_errors, gtk.Label('Errors'))        
                                                                
    def highlight_code(self, widget):
        # Remember which childs in the index are expanded
        exp_functions = self.treeview_index.row_expanded(self.treestore_index.get_path(self.treestore_index_functions))
        exp_data = self.treeview_index.row_expanded(self.treestore_index.get_path(self.treestore_index_data))
        exp_reducs = self.treeview_index.row_expanded(self.treestore_index.get_path(self.treestore_index_reducs))
        exp_processes = self.treeview_index.row_expanded(self.treestore_index.get_path(self.treestore_index_processes))

        self.reset_messages()        
        self.reset_tags()
        self.reset_index()

        # Hide parse error
        self.eventbox_parse_error.hide()                    
        self.label_parse_error.set_text('')
        self.buffer.remove_tag(self.texttag_parse_error, self.buffer.get_start_iter(), self.buffer.get_end_iter())
                            
        try:
            self.parser = Parser()
            (start, end) = self.buffer.get_bounds()
            chars = self.buffer.get_text(start, end, False)
            self.parser.set_text(chars)
            self.parser.parse()
    
            for token in self.parser.functions.values():
                name = token.ident.name

                if isinstance(token, DataToken):
                    self.treestore_index.append(self.treestore_index_data, [name])                
                elif isinstance(token, FunToken):
                    if token.private:
                        name = '*' + name
                    self.treestore_index.append(self.treestore_index_functions, [name])
                elif isinstance(token, ReducToken):
                    if token.private:
                        name = '*' + name                    
                    self.treestore_index.append(self.treestore_index_reducs, [name])
    
            for token in self.parser.processes.values():
                self.treestore_index.append(self.treestore_index_processes, [token.ident.name])
        
            if exp_functions: self.treeview_index.expand_row(self.treestore_index.get_path(self.treestore_index_functions), False)
            if exp_data: self.treeview_index.expand_row(self.treestore_index.get_path(self.treestore_index_data), False)
            if exp_reducs: self.treeview_index.expand_row(self.treestore_index.get_path(self.treestore_index_reducs), False)
            if exp_processes: self.treeview_index.expand_row(self.treestore_index.get_path(self.treestore_index_processes), False)        
        except ParseException as e:
            # Get iter at the position of the exception
            iter = self.buffer.get_iter_at_offset(e.pos)
            
            # Show parse error in editor
            self.buffer.apply_tag(self.texttag_parse_error, self.buffer.get_iter_at_line(iter.get_line()), iter)
            
            # Show text of the parse error
            self.label_parse_error.set_text(e.msg)
            self.eventbox_parse_error.show()
        except ParseDeclarationException as e:
            # Get iter at the start and end position of the exception
            iter_start = self.buffer.get_iter_at_offset(e.start)
            iter_end = self.buffer.get_iter_at_offset(e.end)

            # Show parse error in editor
            self.buffer.apply_tag(self.texttag_parse_error, iter_end, iter_start)

            # Show text of the parse error                        
            self.label_parse_error.set_text(e.msg)
            self.eventbox_parse_error.show()            

    def select_context(self, selection):
        (model, iter) = selection.get_selected()
        
        if not iter == None:
            self.do_insert(model.get_value(iter, 0))
            self.treeview_context.hide_all()

    def select_index(self, selection):
        buffer = self.view.get_buffer()

        (model, iter) = selection.get_selected()
                
        if not iter == None:        
            parent = model.iter_parent(iter)
    
            self.treeview_warnings.get_selection().unselect_all()
            self.treeview_errors.get_selection().unselect_all()        

            # Check if a child is selected and not one of the categories
            if not parent == None:
                if (model.get_value(parent, 0) == 'Functions') | (model.get_value(parent, 0) == 'Data') | (model.get_value(parent, 0) == 'Reductions'):
                    self.reset_tags()
                    name = model.get_value(iter, 0)
                    function = self.parser.functions[name]
                    self.highlight_function(function)

                if (model.get_value(parent, 0) == 'Processes'):
                    self.reset_tags()
                    process = model.get_value(iter, 0)
                    token = self.parser.processes[process]
                    self.highlight_process(token)
        
    def update_context_highlighting(self, buffer, iter, textmark):
        # Only update context highlighting if the changed mark has the name insert
        # This mark normally only changes if the cursor is moved
        if not textmark.get_name() == 'insert': return
        
        # Reset all highlighting tags
        self.reset_tags()        
        # Reset all selection (index, warnings and errors)
        self.reset_selections()

        # We will not highlight when text is selected
        if buffer.get_selection_bounds() == ():
            token = self.parser.functions_int.get(iter.get_offset())
            if not token == None:
                self.highlight_function(token)
                return

            token = self.parser.processes_int.get(iter.get_offset())
            if not token == None:
                self.highlight_process(token)
                return

            token = self.parser.names_int.get(iter.get_offset())
            if not token == None:
                self.highlight_name(token)
                return
            
            token = self.parser.reduc_vars_int.get(iter.get_offset())
            if not token == None:
                self.highlight_var(token)
                return
            
    def highlight_function(self, function):
        # Highlight the declaration of the function
        self.buffer.apply_tag(self.texttag_declaration, self.buffer.get_iter_at_offset(function.ident.start), self.buffer.get_iter_at_offset(function.ident.end))

        # Highlight all other occurences of the function
        for ref in function.ident.refs:
            self.buffer.apply_tag(self.texttag_reference, self.buffer.get_iter_at_offset(ref.start), self.buffer.get_iter_at_offset(ref.end))

    def highlight_process(self, process):
        # Highlight the declaration of the process
        self.buffer.apply_tag(self.texttag_declaration, self.buffer.get_iter_at_offset(process.ident.start), self.buffer.get_iter_at_offset(process.ident.end))
        
        # Highlight all other occurences of the process
        for ref in process.ident.refs:
            self.buffer.apply_tag(self.texttag_reference, self.buffer.get_iter_at_offset(ref.start), self.buffer.get_iter_at_offset(ref.end))

    def highlight_name(self, name):
        # Highlight the declaration of the name
        self.buffer.apply_tag(self.texttag_declaration, self.buffer.get_iter_at_offset(name.start), self.buffer.get_iter_at_offset(name.end))
        
        # Highlight all other occurences of the name
        for ref in name.refs:
            self.buffer.apply_tag(self.texttag_reference, self.buffer.get_iter_at_offset(ref.start), self.buffer.get_iter_at_offset(ref.end))

    def highlight_var(self, var):
        # Highlight the declaration of the variables
        self.buffer.apply_tag(self.texttag_reference, self.buffer.get_iter_at_offset(var.start), self.buffer.get_iter_at_offset(var.end))
        
        # Highlight all other occurences of the variable
        for ref in var.refs:
            self.buffer.apply_tag(self.texttag_reference, self.buffer.get_iter_at_offset(ref.start), self.buffer.get_iter_at_offset(ref.end))

    def remove_children(self, parent):
        child = self.treestore_index.iter_children(parent)
        while not child == None:
            self.treestore_index.remove(child)
            child = self.treestore_index.iter_children(parent)




class Buffer(gtksourceview2.Buffer):
    def __init__(self):
        gtksourceview2.Buffer.__init__(self)
        self.filename = None

    def save(self, filename = None):
        #TODO create backup file
        update_modified = False
        
        if filename == None:
            filename = self.filename
            update_modified = True
            
        result = False
        start, end = self.get_bounds()
        chars = self.get_text(start, end, False)
        try:
            file = open(filename, "w")
            file.write(chars)
            file.close()
            result = True
            
            if update_modified:
                self.set_modified(False)
        except IOError as e:
            err = "Error writing to '%s': %s" % (filename, e.errmsg)
            dialog = gtk.MessageDialog(None, gtk.DIALOG_MODAL,
				      gtk.MESSAGE_INFO,
				      gtk.BUTTONS_OK, err);
            dialog.run()
            dialog.destroy()

        return result

    def load(self):
        try:
            f = open(self.filename, "r")
        except IOError as e:
            err = "Cannot open file '%s': %s" % (self.filename, e.errmsg)
            dialog = gtk.MessageDialog(None, gtk.DIALOG_MODAL,
                                       gtk.MESSAGE_INFO,
                                       gtk.BUTTONS_OK, err);
            dialog.run()
            dialog.destroy()
            return False
  
        buf = f.read()
        f.close()
        self.begin_not_undoable_action()
        self.set_text(buf)
        self.end_not_undoable_action()
        self.set_modified(False)
        return True

    def clear(self):
        self.begin_not_undoable_action()
        self.set_text('')
        self.end_not_undoable_action()
        self.set_modified(False)

    def set_filename(self, filename):
        self.filename = filename

    def get_filename(self):
        return self.filename


# If the program is run directly or passed as an argument to the python
# interpreter then create an instance and show it
if __name__ == "__main__":
    editor = Editor()
    editor.main()


